-- Copyright 1999-2000, Tom Moran, anyone may use for any purpose.
-- This program is offered for educational purposes and has no
-- warranties of any kind.
--
-- This is a simple web server, written in Ada.  It is functionally
-- similar to the Java "A Simple Multithreaded Web Server" at
-- http://developer.java.sun.com/developer/technicalArticles/Networking/Webserver/index.html
-- This uses parts of CLAW (Class Library Ada Windows).  It uses
-- only parts present in the $0 introductory version of CLAW
-- available via http://www.rrsoftware.com, except for packages
-- Claw.Sockets, Claw.Directories, and Claw.Time.  Simplified versions
-- of those will soon be added to the $0 introductory CLAW, allowing
-- the experimenter to download for $0 all the source code needed for
-- a simple web server.

with Ada.Calendar,
     Ada.Characters.Handling,
     Ada.Exceptions,
     Ada.Streams,
     Ada.Streams.Stream_IO,
     Ada.Strings,
     Ada.Strings.Fixed,
     Ada.Strings.Maps,
     Ada.Text_IO,
     Claw.Sockets,
     Claw.Directories,
     Claw.Time;
procedure smplsrvr is
  use type Ada.Calendar.Time;

  Log_File_Name : constant String := "smplsrvr.log";  -- CHANGE AS DESIRED.

  Public_Path : constant String := "r:\public\"; -- CHANGE AS DESIRED.
  -- prepended to any request, so those are the only files client can see.

  Request_Patience : constant Duration := 5.0;   -- CHANGE AS DESIRED.
  -- Once a client has arrived, it shouldn't take long to read his request.

  Server_Port : constant Claw.Sockets.Port_Type := Claw.Sockets.http_Port;
  Path_Separator : constant character := '\';

  function Time_Image(T : in Ada.Calendar.Time := Ada.Calendar.Clock;
        Date_Format : Claw.Time.Date_Format_Type := Claw.Time.Short_Month_Day_First;
        Day_of_Week_Format : Claw.Time.Day_of_Week_Format_Type := Claw.Time.Short;
        Time_Format : Claw.Time.Time_Format_Type := Claw.Time.Normal;
        Date_Separator : Character := '/') return String renames
        Claw.Time.Time_Image;

  protected Log is  -- one at a time can write in log
    procedure Note(Msg : in String);
    procedure Close;
  private
    Log_File : Ada.Text_IO.File_Type;
  end Log;

  protected body Log is
    procedure Note(Msg : in String) is
    begin
      if not Ada.Text_IO.Is_Open(Log_File) then
        Ada.Text_IO.Create(Log_File, Ada.Text_IO.Out_File, Log_File_Name);
      end if;
        Ada.Text_IO.Put_Line(Log_File, Msg);
    end Note;
    procedure Close is
    begin
      if Ada.Text_IO.Is_Open(Log_File) then Ada.Text_IO.Close(Log_File);end if;
    end Close;
  end Log;

  function Content_Label(Ext : in String) return String is
  begin
    if Ext'length < 2 then return "content/unknown";end if;
    if Ext = ".htm" or Ext = ".html" then return "text/html";end if;
    if Ada.Strings.Fixed.Index(".text .c .cc .c++ .h .pl .txt .java", Ext)
         /= 0 then
      return "text/plain";
    end if;
    if Ext = ".exe" or Ext = ".uu" then return "application/octet-stream";end if;
    if Ext = ".ps" then return "application/postscript";end if;
    if Ext = ".zip" then return "application/zip";end if;
    if Ext = ".sh" then return "application/x-shar";end if;
    if Ext = ".tar" then return "application/x-tar";end if;
    if Ext = ".snd" or Ext = ".au" then return "audio/basic";end if;
    if Ext = ".wav" then return "audio/x-wav";end if;
    if Ext = ".gif" then return "image/gif";end if;
    if Ext = ".jpg" or Ext = ".jpeg" then return "image/jpeg";end if;
    return "unknown/unknown";
  end Content_Label;

  function Normalize(Name : in String) return String is
  -- Replaces '/' by Path_Separator,
  -- drops any initial or trailing Path_Separator.
    Fixed : String := Name;
    First : Natural := Fixed'first;
    Last : Natural := Fixed'last;
  begin
    if Fixed'length = 0 then return "";end if;
    for i in Fixed'range loop
      if Fixed(i) = '/' or Fixed(i) = '\' then
        Fixed(i) := Path_Separator;
      end if;
    end loop;
    if Fixed(First) = Path_Separator then First := First+1;end if;
    if Fixed(Last) = Path_Separator then Last := Last-1;end if;
    if Last >= First then
      return Fixed(First .. Last);
    else
      return "";
    end if;
  end Normalize;

  procedure Send_Header(Client : in out Claw.Sockets.Socket_Type;
                        F : in Claw.Directories.Directory_Entry_Type) is
  begin
    if Claw.Directories.Is_Directory(F) then
      Claw.Sockets.Put_Line(Client, "Content-type: text/html");
    else
      Claw.Sockets.Put_Line(Client, "Content-length:"
          & Claw.DWord'image(Claw.Directories.Size_Of(F)));
      Claw.Sockets.Put_Line(Client, "Last-Modified: "
          & Time_Image (Claw.Time.To_UTC_Time (
                Claw.Directories.Last_Write_Time_Of (F)))
          & " GMT");
      declare
        Long_Name : constant String  -- only care about .xxx part
          := Ada.Characters.Handling.To_Lower(Claw.Directories.File_Name_Of(F));
        Dot_Index : constant Natural
          := Ada.Strings.Fixed.Index(Long_Name, ".", Ada.Strings.Backward);
      begin
        if Dot_Index = 0 then
          Claw.Sockets.Put_Line(Client, "Content-type: unknown/unknown");
        else
          Claw.Sockets.Put_Line(Client, "Content-type: "
              & Content_Label(Long_Name(Dot_Index .. Long_Name'last)));
        end if;
      end;
    end if;
    Claw.Sockets.Put_Line(Client, "");
  end Send_Header;

  procedure Send_File(Client : in out Claw.Sockets.Socket_Type;
                      Full_Name : in String;
                      F : in Claw.Directories.Directory_Entry_Type) is
    use Ada.Streams, Ada.Streams.Stream_IO;
    Buffer : Stream_Element_Array(1 .. Stream_Element_Count(
        Claw.Directories.Size_Of(F)));
    Last : Stream_Element_Offset;
    String_Buffer : String(1 .. Buffer'length); -- no Stream IO in demo CLAW
    File : File_Type;
  begin
    if Buffer'length = 0 then return;end if;
    Open(File, In_File, Full_Name);
    Read(File, Buffer, Last);
    for i in String_Buffer'range loop
      String_Buffer(i) := Character'val(Stream_Element'pos(Buffer(Stream_Element_Offset(i))));
    end loop;
    Claw.Sockets.Put(Client, String_Buffer); -- no Ada.Streams IO in demo CLAW
    Close(File);
  exception
    when End_Error => Close(File);
    when Name_Error => Log.Note("Missing File:" & Claw.Directories.File_Name_Of(F));
    when Err:others =>
      if Is_Open(File) then Close(File);end if;
      Log.Note("send file err:" & ada.exceptions.exception_information(err));
  end Send_File;

  type Method_Type is (GET, HEAD, Unknown);

  subtype Status_Codes is String(1 .. 3);
  HTTP_OK: constant Status_Codes := "200";
  HTTP_NOT_FOUND: constant Status_Codes := "404";
  HTTP_BAD_METHOD: constant Status_Codes := "405";

  procedure Respond(Client : in out Claw.Sockets.Socket_Type;
                    Method : in Method_Type;    -- eg Get
                    Request_URI : in String) is  -- eg a/b/c.htm

    procedure List_File_Name(F    : in     Claw.Directories.Directory_Entry_Type;
                             Quit : in out Boolean) is
      File_Name : constant String := Claw.Directories.File_Name_Of(F);
    begin
      if File_Name(File_Name'first) /= '.' then -- ignore . and .. entries
        if Claw.Directories.Is_Directory(F) then
          Claw.Sockets.Put_Line(Client, "<A HREF=" & '"'
            & File_Name & "/" & '"' & '>' & File_Name & "/</A><BR>");
        else
          Claw.Sockets.Put_Line(Client, "<A HREF=" & '"'
            & File_Name & '"' & '>' & File_Name & "</A><BR>");
        end if;
      end if;
    end List_File_Name;
    procedure List_Directory is new Claw.Directories.For_Every_Directory_Entry(List_File_Name);

    Result_Sent : Boolean := False;

    procedure Find_And_Do_File(Full_Name : in String) is

      procedure Send_Result(F    : in     Claw.Directories.Directory_Entry_Type;
                            Quit : in out Boolean) is
      -- Send header info.  If a GET, also send file or directory content.
      begin
        Quit := True; -- Just one file should match the name
        if (Request_URI(Request_URI'last) /= '/')    -- asked for file,
        and Claw.Directories.Is_Directory(F) then    -- found directory
          return; -- ignore the directory (to not confuse browsers).
        end if;
        Result_Sent := True;
        Claw.Sockets.Put_Line(Client, "HTTP/1.0 " & HTTP_OK  & " OK");
        Claw.Sockets.Put_Line(Client, "Server: Simple Ada");
        Claw.Sockets.Put_Line(Client, "Date: " & Time_Image);
        Send_Header(Client, F);
        if Method = Get then
          if Claw.Directories.Is_Directory(F) then
            Claw.Sockets.Put_Line(Client, "<TITLE>Directory listing</TITLE><P>");
            Claw.Sockets.Put_Line(Client, "<A HREF=" & '"' & ".." & '"'
              & ">Parent Directory</A><BR>");
            List_Directory(Full_Name & Path_Separator & "*.*",
                           Ignore_Not_Found_Error => True);
            Claw.Sockets.Put_Line(Client, "<P><HR><BR><I>" & Time_Image & "</I>");
          else
            Send_File(Client, Full_Name, F);
          end if;
        end if;
      end Send_Result;

      procedure Handle_The_File is new
                Claw.Directories.For_Every_Directory_Entry(Send_Result);
    begin
      Handle_The_File(Full_Name, Ignore_Not_Found_Error => True);
    end Find_And_Do_File;

    Canonical_Name : constant String := Normalize(Request_URI);

  begin
    if Ada.Strings.Fixed.Index(Canonical_Name, "?") = 0
    and then Ada.Strings.Fixed.Index(Canonical_Name, "*") = 0
    and then Ada.Strings.Fixed.Index(Canonical_Name, "..") = 0 then
      -- first try Name/index.html
      if Canonical_Name'length = 0 then -- don't list root directory
        Find_And_Do_File(Public_Path & "index.html");
      else
        Find_And_Do_File(Public_Path & Canonical_Name & Path_Separator & "index.html");
        if not Result_Sent then -- try Name alone
          Find_And_Do_File(Public_Path & Canonical_Name);
        end if;
      end if;
    end if;
    if not Result_Sent then
      Claw.Sockets.Put_Line(Client, "HTTP/1.0 " & HTTP_NOT_FOUND & " not found");
      Claw.Sockets.Put_Line(Client, "Server: Simple Ada");
      Claw.Sockets.Put_Line(Client, "Date: " & Time_Image);
    end if;
    declare
      Client_Address : Claw.Sockets.Network_Address_Type;
      Client_Port : Claw.Sockets.Port_Type;
    begin
      Claw.Sockets.Get_Peer(Client, Client_Address, Client_Port);
      Log.Note("From " & Claw.Sockets.String_Of(Client_Address)
        & ": " & Method_Type'image(Method) & " " & Request_URI
        & " found? " & Boolean'image(Result_Sent));
    exception
      when others => null; -- forget it (perhaps browser has disconnected?)
    end;
  exception
    when Err:others =>
      Log.Note("respond err:" & ada.exceptions.exception_information(err));
  end Respond;

  task type Worker(ID : Character) is
    entry Get_Job(Server : in out Claw.Sockets.Server_Type);
    -- Call with Server.  Worker will then open his own socket to the client.
    entry Quit;
  end Worker;
  for Worker'storage_size use 16384;

  type A_Worker_Handle is access Worker;

  Roster : array(Character range 'A' .. 'F') of A_Worker_Handle;

  task body Worker is
    Line : String(1 .. 2048);
    Last : Natural := 0;
    EOL_Chars : constant Ada.Strings.Maps.Character_Set
      := Ada.Strings.Maps.To_Set(ascii.cr & ascii.lf);
    Method : Method_Type;
    Name_First, After_Name : Natural;
    Client : Claw.Sockets.Socket_Type;
  begin
    Log.Note("New Worker " & ID);
    loop
      begin
        select
          accept Get_Job(Server : in out Claw.Sockets.Server_Type) do
            Claw.Sockets.Greet(Server, Client);
          end Get_Job;
          Last := 0;
          while Last < Line'last loop
            Claw.Sockets.Get(Client, Request_Patience, Line(Last+1 .. Line'last), Last);
            exit when Last >= Line'first
              and then Ada.Strings.Fixed.Index(Line(Line'first .. Last), EOL_Chars) /= 0;
          end loop;
          -- proper format is   Method SP Request-URI SP HTTP-Version CRLF
          if Last >= 7 and then Line(Line'first .. Line'first+3) = "GET " then
            Method := Get;
            Name_First := Line'first+4;
          elsif Last >= 8 and then Line(Line'first .. Line'first+4) = "HEAD " then
            Method := Head;
            Name_First := Line'first+5;
          else
            Method := Unknown;
          end if;
          After_Name := Ada.Strings.Fixed.Index(Line(Line'first+5 .. Last)," ");
          if After_Name = 0 then Method := Unknown;end if;
          if Method /= Unknown then
            Log.Note(ID & " Method " & Method_Type'image(Method) & ' '
              & Line(Name_First .. After_Name-1));
            Respond(Client, Method, Line(Name_First .. After_Name-1));
          else
            Log.Note("unknown method:" & Line(Line'first .. Last));
            Claw.Sockets.Put_Line(Client, "HTTP/1.0 " & HTTP_BAD_METHOD
                   & " unsupported method requested: "
                   & Line(Line'first .. integer'min(Last, Line'first+3)));
          end if;
          Claw.Sockets.Close(Client);
        or
          accept Quit; exit;
        end select;
      exception
        when Err:others =>
          Log.Note("worker err:" & ada.exceptions.exception_information(err));
      end;
    end loop;
  end Worker;

  procedure Stop_Workers is
  begin
    for i in Roster'range loop
      if Roster(i) /= null then
        begin
          select
            Roster(i).Quit;
          or delay 2.0;
          end select;
        exception
          when Tasking_Error => null;
        end;
      end if;
    end loop;
  end Stop_Workers;

  procedure Serve(End_Time : in Ada.Calendar.Time) is
    use Ada.Calendar;
    Server : Claw.Sockets.Server_Type;
    Ready : Boolean;
  begin
    Claw.Sockets.Open(Server, Port=>Server_Port);
    loop
      delay 0.1;
      exit when Clock > End_Time;
      Claw.Sockets.Readable(Server, Ready);
      if Ready then
        Find_A_Worker:
        for Remaining_Tries in reverse 0 .. 5 loop
          for i in Roster'range loop
            if Roster(i) = null then  -- may start a new worker
              Roster(i) := new Worker(i);
              Roster(i).Get_Job(Server);
              exit Find_A_Worker;
            else                      -- or try to reuse the worker
              select
                Roster(i).Get_Job(Server);
                exit Find_A_Worker;
              or delay 0.2;
              end select;
            end if;
          end loop;
          -- No available workers!  Should we try again?
          if Remaining_Tries = 0 then
            Log.Note("no worker!");
          end if;
        end loop Find_A_Worker;
      end if;
    end loop;
    Claw.Sockets.Close(Server);
    Stop_Workers;
  exception
    when egad:others =>
      Log.Note("Server error:" & ada.exceptions.exception_information(egad));
      Stop_Workers;
  end Serve;

begin
  Log.Note("Starting simple HTML server");
  Serve(Ada.Calendar.Clock+120.0);  -- run for 2 minutes, then quit.
  Log.Note("Stopping simple HTML server");
  Log.Close;
exception
  when oops:others=>
    Log.Note("Error:" & ada.exceptions.exception_information(oops));
    Log.Close;
end smplsrvr;
